//-----------------------------------------------------------------------------
// Copyright (c) 2006-2008 dhpoware. All Rights Reserved.
//
// Permission is hereby granted, free of charge, to any person obtaining a
// copy of this software and associated documentation files (the "Software"),
// to deal in the Software without restriction, including without limitation
// the rights to use, copy, modify, merge, publish, distribute, sublicense,
// and/or sell copies of the Software, and to permit persons to whom the
// Software is furnished to do so, subject to the following conditions:
//
// The above copyright notice and this permission notice shall be included in
// all copies or substantial portions of the Software.
//
// THE SOFTWARE IS PROVIDED "AS IS", WITHOUT WARRANTY OF ANY KIND, EXPRESS
// OR IMPLIED, INCLUDING BUT NOT LIMITED TO THE WARRANTIES OF MERCHANTABILITY,
// FITNESS FOR A PARTICULAR PURPOSE AND NONINFRINGEMENT. IN NO EVENT SHALL THE
// AUTHORS OR COPYRIGHT HOLDERS BE LIABLE FOR ANY CLAIM, DAMAGES OR OTHER
// LIABILITY, WHETHER IN AN ACTION OF CONTRACT, TORT OR OTHERWISE, ARISING
// FROM, OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS
// IN THE SOFTWARE.
//-----------------------------------------------------------------------------

#if !defined(SMART_PTR_H)
#define SMART_PTR_H

//-----------------------------------------------------------------------------
// Any class derived from the Referable abstract base class can use the
// smart pointer template class (defined below).
//-----------------------------------------------------------------------------

class Referable
{
public:
    Referable() : m_refCount(0) {}
    virtual ~Referable() {}

    int addRef()
    { return ++m_refCount; }

    int releaseRef()
    { return --m_refCount; }
        
    int refCount() const
    { return m_refCount; }

private:
    int m_refCount;
};

//-----------------------------------------------------------------------------
// An intrusive reference counting smart pointer template class. This smart
// pointer class only works with objects derived from the Referable abstract
// class (defined above). Based on ideas presented in Andrei Alexandrescu's
// "Modern C++ Design" and Scott Meyers' "More Effective C++".
//-----------------------------------------------------------------------------

template <typename ReferableType>
class SmartPtr
{
public:
    SmartPtr(ReferableType *pReferable) : m_pReferable(0)
    { addRef(pReferable); }

    SmartPtr(const SmartPtr &rhs) : m_pReferable(0)
    { addRef(rhs.m_pReferable); }

    SmartPtr() : m_pReferable(0) {}
    
    ~SmartPtr()
    { release(); }

    static ReferableType *getImpl(const SmartPtr &sp)
    { return sp.m_pReferable; }

    template <typename T>
    operator SmartPtr<T>()
    { return SmartPtr<T>(m_pReferable); }

    ReferableType &operator*() const
    { return *m_pReferable; }

    ReferableType *operator->() const
    { return m_pReferable; }

    SmartPtr &operator=(const SmartPtr &rhs)
    {
        if (this != &rhs)
            addRef(rhs.m_pReferable);

        return *this;
    }

    SmartPtr &operator=(ReferableType *pReferable)
    {
        addRef(pReferable);
        return *this;
    }

    bool operator!() const
    { return m_pReferable == 0; }

    friend bool operator==(const SmartPtr &lhs, const ReferableType *rhs)
    { return lhs.m_pReferable == rhs; }

    friend bool operator==(const ReferableType *lhs, const SmartPtr &rhs)
    { return lhs == rhs.m_pReferable; }

    friend bool operator!=(const SmartPtr &lhs, const ReferableType *rhs)
    { return lhs.m_pReferable != rhs; }

    friend bool operator!=(const ReferableType *lhs, const SmartPtr &rhs)
    { return lhs != rhs.m_pReferable; }

    template <typename T>
    friend bool operator==(const SmartPtr &lhs, const T *rhs)
    { return lhs.m_pReferable == rhs; }

    template <typename T>
    friend bool operator==(const T *lhs, const SmartPtr &rhs)
    { return lhs == rhs.m_pReferable; }

    template <typename T>
    friend bool operator!=(const SmartPtr &lhs, const T *rhs)
    { return lhs.m_pReferable != rhs; }

    template <typename T>
    friend bool operator!=(const T *lhs, const SmartPtr &rhs)
    { return lhs != rhs.m_pReferable; }

    template <typename T>
    bool operator==(const SmartPtr<T> &rhs) const
    { return m_pReferable == rhs.m_pReferable; }

    template <typename T>
    bool operator!=(const SmartPtr<T> &rhs) const
    { return m_pReferable != rhs.m_pReferable; }

private:
    class NotNullTester
    {
        void operator delete(void*);
    };

public:
    operator NotNullTester*() const
    {
        if (!m_pReferable)
            return 0;

        static NotNullTester n;
        return &n;
    }

private:
    void addRef(ReferableType *pReferable)
    {
        if (!m_pReferable)
        {
            m_pReferable = pReferable;

            if (m_pReferable)
                m_pReferable->addRef();
        }
        else
        {
            if (m_pReferable != pReferable)
            {
                release();
                m_pReferable = pReferable;

                if (m_pReferable)
                    m_pReferable->addRef();
            }
        }
    }

    void release()
    {
        if (m_pReferable)
        {
            if (m_pReferable->releaseRef() <= 0)
            {
                delete m_pReferable;
                m_pReferable = 0;
            }
        }
    }

    ReferableType *m_pReferable;
};

#endif